@@ -13,17 +13,18 @@ import stat
 import tempfile
 import textwrap
 import uuid
-from collections import Iterable  # pylint: disable=unused-import
+
 from io import open
 from typing import (Any, Callable, Dict, Generator, Iterator, List,
                     Mapping, MutableMapping, MutableSequence, Optional, Set, Tuple,
-                    Union, cast)
+                    Type, Union, cast)
 
 from pkg_resources import resource_stream
 from rdflib import Graph  # pylint: disable=unused-import
 from ruamel.yaml.comments import CommentedMap, CommentedSeq
 from six import PY3, iteritems, itervalues, string_types, with_metaclass
 from six.moves import urllib
+from future.utils import raise_from
 from typing_extensions import (TYPE_CHECKING,  # pylint: disable=unused-import
                                Text)
 from schema_salad import schema, validate
@@ -46,6 +47,8 @@ from .stdfsaccess import StdFsAccess
 from .utils import (DEFAULT_TMP_PREFIX, aslist, cmp_like_py2,
                     copytree_with_merge, onWindows, random_outdir)
 from .validate_js import validate_js_expressions
+from .update import INTERNAL_VERSION
+
 try:
     from os import scandir  # type: ignore
 except ImportError:
@@ -53,14 +56,19 @@ except ImportError:
 if TYPE_CHECKING:
     from .provenance import ProvenanceProfile  # pylint: disable=unused-import
 
+if PY3:
+    from collections.abc import Iterable # only works on python 3.3+
+else:
+    from collections import Iterable  # pylint: disable=unused-import
 
 class LogAsDebugFilter(logging.Filter):
     def __init__(self, name, parent):  # type: (Text, logging.Logger) -> None
+        """Initialize."""
         name = str(name)
         super(LogAsDebugFilter, self).__init__(name)
         self.parent = parent
 
-    def filter(self, record):
+    def filter(self, record):  # type: (logging.LogRecord) -> bool
         return self.parent.isEnabledFor(logging.DEBUG)
 
 
@@ -79,9 +87,11 @@ supportedProcessRequirements = ["DockerRequirement",
                                 "StepInputExpressionRequirement",
                                 "ResourceRequirement",
                                 "InitialWorkDirRequirement",
-                                "TimeLimit",
+                                "ToolTimeLimit",
                                 "WorkReuse",
                                 "NetworkAccess",
+                                "InplaceUpdateRequirement",
+                                "LoadListingRequirement",
                                 "http://commonwl.org/cwltool#TimeLimit",
                                 "http://commonwl.org/cwltool#WorkReuse",
                                 "http://commonwl.org/cwltool#NetworkAccess",
@@ -150,7 +160,7 @@ def get_schema(version):
     if version in SCHEMA_CACHE:
         return SCHEMA_CACHE[version]
 
-    cache = {}  # type: Dict[Text, Union[bytes, Text]]
+    cache = {}  # type: Dict[Text, Any]
     version = version.split("#")[-1]
     if '.dev' in version:
         version = ".".join(version.split(".")[:-1])
@@ -209,17 +219,17 @@ def checkRequirements(rec, supported_process_requirements):
 
 
 def stage_files(pathmapper,             # type: PathMapper
-                stage_func=None,        # type: Callable[..., Any]
+                stage_func=None,        # type: Optional[Callable[..., Any]]
                 ignore_writable=False,  # type: bool
                 symlink=True,           # type: bool
-                secret_store=None       # type: SecretStore
+                secret_store=None       # type: Optional[SecretStore]
                ):  # type: (...) -> None
     """Link or copy files to their targets. Create them as needed."""
     for key, entry in pathmapper.items():
         if not entry.staged:
             continue
         if not os.path.exists(os.path.dirname(entry.target)):
-            os.makedirs(os.path.dirname(entry.target), 0o0755)
+            os.makedirs(os.path.dirname(entry.target))
         if entry.type in ("File", "Directory") and os.path.exists(entry.resolved):
             if symlink:  # Use symlink func if allowed
                 if onWindows():
@@ -236,13 +246,13 @@ def stage_files(pathmapper,             # type: PathMapper
                 stage_func(entry.resolved, entry.target)
         elif entry.type == "Directory" and not os.path.exists(entry.target) \
                 and entry.resolved.startswith("_:"):
-            os.makedirs(entry.target, 0o0755)
+            os.makedirs(entry.target)
         elif entry.type == "WritableFile" and not ignore_writable:
             shutil.copy(entry.resolved, entry.target)
             ensure_writable(entry.target)
         elif entry.type == "WritableDirectory" and not ignore_writable:
             if entry.resolved.startswith("_:"):
-                os.makedirs(entry.target, 0o0755)
+                os.makedirs(entry.target)
             else:
                 shutil.copytree(entry.resolved, entry.target)
                 ensure_writable(entry.target)
@@ -261,13 +271,13 @@ def stage_files(pathmapper,             # type: PathMapper
                 key, entry.target, entry.target, entry.type, entry.staged)
 
 
-def relocateOutputs(outputObj,             # type: Union[Dict[Text, Any],List[Dict[Text, Any]]]
-                    destination_path,      # type: Text
-                    source_directories,    # type: Set[Text]
-                    action,                # type: Text
-                    fs_access,             # type: StdFsAccess
+def relocateOutputs(outputObj,              # type: Union[Dict[Text, Any], List[Dict[Text, Any]]]
+                    destination_path,       # type: Text
+                    source_directories,     # type: Set[Text]
+                    action,                 # type: Text
+                    fs_access,              # type: StdFsAccess
                     compute_checksum=True,  # type: bool
-                    path_mapper=PathMapper
+                    path_mapper=PathMapper  # type: Type[PathMapper]
                     ):
     # type: (...) -> Union[Dict[Text, Any], List[Dict[Text, Any]]]
     adjustDirObjs(outputObj, functools.partial(get_listing, fs_access, recursive=True))
@@ -289,11 +299,12 @@ def relocateOutputs(outputObj,             # type: Union[Dict[Text, Any],List[Di
                 for dir_entry in _collectDirEntries(sub_obj):
                     yield dir_entry
 
-    def _relocate(src, dst):
+    def _relocate(src, dst):  # type: (Text, Text) -> None
         if src == dst:
             return
 
         # If the source is not contained in source_directories we're not allowed to delete it
+        src = fs_access.realpath(src)
         src_can_deleted = any(os.path.commonprefix([p, src]) == p for p in source_directories)
 
         _action = "move" if action == "move" and src_can_deleted else "copy"
@@ -322,7 +333,7 @@ def relocateOutputs(outputObj,             # type: Union[Dict[Text, Any],List[Di
     pm = path_mapper(outfiles, "", destination_path, separateDirs=False)
     stage_files(pm, stage_func=_relocate, symlink=False)
 
-    def _check_adjust(a_file):
+    def _check_adjust(a_file):  # type: (Dict[Text, Text]) -> Dict[Text, Text]
         a_file["location"] = file_uri(pm.mapper(a_file["location"])[1])
         if "contents" in a_file:
             del a_file["contents"]
@@ -355,7 +366,7 @@ def add_sizes(fsaccess, obj):  # type: (StdFsAccess, Dict[Text, Any]) -> None
         return  # best effort
 
 def fill_in_defaults(inputs,   # type: List[Dict[Text, Text]]
-                     job,      # type: Dict[Text, Union[Dict[Text, Any], List[Any], Text, None]]
+                     job,      # type: Dict[Text, expression.JSON]
                      fsaccess  # type: StdFsAccess
                     ):  # type: (...) -> None
     for e, inp in enumerate(inputs):
@@ -373,9 +384,7 @@ def fill_in_defaults(inputs,   # type: List[Dict[Text, Text]]
 
 def avroize_type(field_type, name_prefix=""):
     # type: (Union[List[Dict[Text, Any]], Dict[Text, Any]], Text) -> Any
-    """
-    adds missing information to a type so that CWL types are valid in schema_salad.
-    """
+    """Add missing information to a type so that CWL types are valid."""
     if isinstance(field_type, MutableSequence):
         for field in field_type:
             avroize_type(field, name_prefix)
@@ -411,11 +420,11 @@ _VAR_SPOOL_ERROR = textwrap.dedent(
     """)
 
 
-def var_spool_cwl_detector(obj,           # type: Union[MutableMapping, List, Text]
+def var_spool_cwl_detector(obj,           # type: Union[MutableMapping[Text, Text], List[Dict[Text, Any]], Text]
                            item=None,     # type: Optional[Any]
                            obj_key=None,  # type: Optional[Any]
                           ):              # type: (...)->bool
-    """ Detects any textual reference to /var/spool/cwl. """
+    """Detect any textual reference to /var/spool/cwl."""
     r = False
     if isinstance(obj, string_types):
         if "var/spool/cwl" in obj and obj_key != "dockerOutputDirectory":
@@ -424,11 +433,11 @@ def var_spool_cwl_detector(obj,           # type: Union[MutableMapping, List, Te
                     _VAR_SPOOL_ERROR.format(obj)))
             r = True
     elif isinstance(obj, MutableMapping):
-        for key, value in iteritems(obj):
-            r = var_spool_cwl_detector(value, obj, key) or r
+        for mkey, mvalue in iteritems(obj):
+            r = var_spool_cwl_detector(mvalue, obj, mkey) or r
     elif isinstance(obj, MutableSequence):
-        for key, value in enumerate(obj):
-            r = var_spool_cwl_detector(value, obj, key) or r
+        for lkey, lvalue in enumerate(obj):
+            r = var_spool_cwl_detector(lvalue, obj, lkey) or r
     return r
 
 def eval_resource(builder, resource_req):  # type: (Builder, Text) -> Any
@@ -436,6 +445,7 @@ def eval_resource(builder, resource_req):  # type: (Builder, Text) -> Any
         return builder.do_eval(resource_req)
     return resource_req
 
+
 # Threshold where the "too many files" warning kicks in
 FILE_COUNT_WARNING = 5000
 
@@ -444,6 +454,7 @@ class Process(with_metaclass(abc.ABCMeta, HasReqsHints)):
                  toolpath_object,      # type: MutableMapping[Text, Any]
                  loadingContext        # type: LoadingContext
                 ):  # type: (...) -> None
+        """Build a Process object from the provided dictionary."""
         self.metadata = getdefault(loadingContext.metadata, {})  # type: Dict[Text,Any]
         self.provenance_object = None  # type: Optional[ProvenanceProfile]
         self.parent_wf = None          # type: Optional[ProvenanceProfile]
@@ -462,6 +473,8 @@ class Process(with_metaclass(abc.ABCMeta, HasReqsHints)):
         self.tool = toolpath_object
         self.requirements = copy.deepcopy(getdefault(loadingContext.requirements, []))
         self.requirements.extend(self.tool.get("requirements", []))
+        if "id" not in self.tool:
+            self.tool["id"] = "_:" + Text(uuid.uuid4())
         self.requirements.extend(get_overrides(getdefault(loadingContext.overrides_list, []),
                                                self.tool["id"]).get("requirements", []))
         self.hints = copy.deepcopy(getdefault(loadingContext.hints, []))
@@ -544,7 +557,7 @@ class Process(with_metaclass(abc.ABCMeta, HasReqsHints)):
                     _logger.error(
                         "Failed to read options file %s",
                         loadingContext.js_hint_options_file)
-                    raise err
+                    raise
             else:
                 validate_js_options = None
             if self.doc_schema is not None:
@@ -576,18 +589,22 @@ class Process(with_metaclass(abc.ABCMeta, HasReqsHints)):
     def _init_job(self, joborder, runtime_context):
         # type: (Mapping[Text, Text], RuntimeContext) -> Builder
 
-        job = cast(Dict[Text, Union[Dict[Text, Any], List[Any], Text, None]],
-                   copy.deepcopy(joborder))
+        if self.metadata.get("cwlVersion") != INTERNAL_VERSION:
+            raise WorkflowException("Process object loaded with version '%s', must update to '%s' in order to execute." % (
+                self.metadata.get("cwlVersion"), INTERNAL_VERSION))
+
+        job = cast(Dict[Text, expression.JSON], copy.deepcopy(joborder))
 
         make_fs_access = getdefault(runtime_context.make_fs_access, StdFsAccess)
         fs_access = make_fs_access(runtime_context.basedir)
 
         load_listing_req, _ = self.get_requirement(
-            "http://commonwl.org/cwltool#LoadListingRequirement")
+            "LoadListingRequirement")
+
         if load_listing_req is not None:
             load_listing = load_listing_req.get("loadListing")
         else:
-            load_listing = "deep_listing"   # will default to "no_listing" in CWL v1.1
+            load_listing = "no_listing"
 
         # Validate job order
         try:
@@ -606,13 +623,14 @@ class Process(with_metaclass(abc.ABCMeta, HasReqsHints)):
 
             visit_class(job, ("File",), functools.partial(add_sizes, fs_access))
 
-            if load_listing == "deep_listing" and load_listing_req is None:
+            if load_listing == "deep_listing":
                 for i, inparm in enumerate(self.tool["inputs"]):
                     k = shortname(inparm["id"])
                     if k not in job:
                         continue
                     v = job[k]
                     dircount = [0]
+
                     def inc(d):  # type: (List[int]) -> None
                         d[0] += 1
                     visit_class(v, ("Directory",), lambda x: inc(dircount))
@@ -636,7 +654,7 @@ hints:
 """ % (filecount[0], k))))
 
         except (validate.ValidationException, WorkflowException) as err:
-            raise WorkflowException("Invalid job input record:\n" + Text(err))
+            raise_from(WorkflowException("Invalid job input record:\n" + Text(err)), err)
 
         files = []  # type: List[Dict[Text, Text]]
         bindings = CommentedSeq()
@@ -660,7 +678,7 @@ hints:
                         runtime_context.docker_outdir or random_outdir()
             elif default_docker is not None:
                 outdir = runtime_context.docker_outdir or random_outdir()
-            tmpdir = runtime_context.docker_tmpdir or "/tmp"
+            tmpdir = runtime_context.docker_tmpdir or "/tmp"  # nosec
             stagedir = runtime_context.docker_stagedir or "/var/lib/cwl"
         else:
             outdir = fs_access.realpath(
@@ -714,7 +732,14 @@ hints:
                 if isinstance(arg, MutableMapping):
                     arg = copy.deepcopy(arg)
                     if arg.get("position"):
-                        arg["position"] = [arg["position"], i]
+                        position = arg.get("position")
+                        if isinstance(position, str):  # no need to test the
+                                                       # CWLVersion as the v1.0
+                                                       # schema only allows ints
+                            position = builder.do_eval(position)
+                            if position is None:
+                                position = 0
+                        arg["position"] = [position, i]
                     else:
                         arg["position"] = [0, i]
                     bindings.append(arg)
@@ -760,11 +785,17 @@ hints:
         resourceReq, _ = self.get_requirement("ResourceRequirement")
         if resourceReq is None:
             resourceReq = {}
+        cwl_version = self.metadata.get(
+            "http://commonwl.org/cwltool#original_cwlVersion", None)
+        if cwl_version == "v1.0":
+            ram = 1024
+        else:
+            ram = 256
         request = {
             "coresMin": 1,
             "coresMax": 1,
-            "ramMin": 1024,
-            "ramMax": 1024,
+            "ramMin": ram,
+            "ramMax": ram,
             "tmpdirMin": 1024,
             "tmpdirMax": 1024,
             "outdirMin": 1024,
@@ -806,8 +837,10 @@ hints:
                     validate.validate_ex(
                         avsc_names.get_name(plain_hint["class"], ""),
                         plain_hint, strict=strict)
+                elif r["class"] in ("NetworkAccess", "LoadListingRequirement"):
+                    pass
                 else:
-                    _logger.info(sl.makeError(u"Unknown hint %s" % (r["class"])))
+                    _logger.info(Text(sl.makeError(u"Unknown hint %s" % (r["class"]))))
 
     def visit(self, op):  # type: (Callable[[MutableMapping[Text, Any]], None]) -> None
         op(self.tool)
@@ -825,7 +858,7 @@ hints:
 _names = set()  # type: Set[Text]
 
 
-def uniquename(stem, names=None):  # type: (Text, Set[Text]) -> Text
+def uniquename(stem, names=None):  # type: (Text, Optional[Set[Text]]) -> Text
     global _names
     if names is None:
         names = _names
@@ -948,8 +981,12 @@ def scandeps(base,                          # type: Text
                             base, u, reffields, urlfields, loadref,
                             urljoin=urljoin, nestdirs=nestdirs))
                     else:
-                        sub = loadref(base, u)
                         subid = urljoin(base, u)
+                        basedf, _ = urllib.parse.urldefrag(base)
+                        subiddf, _ = urllib.parse.urldefrag(subid)
+                        if basedf == subiddf:
+                            continue
+                        sub = loadref(base, u)
                         deps = {
                             "class": "File",
                             "location": subid,
@@ -989,9 +1026,9 @@ def scandeps(base,                          # type: Text
     return r
 
 
-def compute_checksums(fs_access, fileobj):
+def compute_checksums(fs_access, fileobj):  # type: (StdFsAccess, Dict[Text, Any]) -> None
     if "checksum" not in fileobj:
-        checksum = hashlib.sha1()
+        checksum = hashlib.sha1()  # nosec
         with fs_access.open(fileobj["location"], "rb") as f:
             contents = f.read(1024 * 1024)
             while contents != b"":
